/* YABRI - Yet Another Bridge for Ethernet
 * File reply.c gives answers to the outside ....
 * (c) Rainer Bawidamann, Carsten Gross 1995
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, see COPYING
 * for details
 */

/* The packet driver interface is....
 * public domain by Russell Nelson, nelson@crynwr.com.  Politeness dictates
 * that you leave this notice intact
 */

static char *reply = "Rainer Bawidamann & Carsten Gross 1996";

/* reply.c
 *  Version 0.1
 */

/* History
 * 6.5.97
 * Corrected some bugs in tftp and udp/tcp packet handling. Should be okay
 * by now. The tx? option for ODI-driver is removed again (won't work)
 *
 * 13.5.96
 * Added this file reply.c to split yabri.c into 2 parts.
 *
 *    This is free software, you can modify, distribute and copy it under
 *    the terms of the GNU General Public License, see COPYING for details
 * take a look at yabri.c for changes at this file
 */

#include <stdio.h>
#include <string.h>
#include <conio.h>
#include <dos.h>
#include <stdlib.h>
#include "pktdrvr.h"
#include "yabri.h"

#undef debug_ping
#define debug_arp
#define DEBUG_TFTP

/* global variables used in reply.c
 */

int blocknumber=1;		/* Data block in tftp connect */
int conn_established=0;         /* tftp connect. established  */
char messagebuffer[2000];       /* messagebuffer for tftp     */
char *output;                   /* Pointer to messagebuffer   */

/* variables imported from yabri.c
 */

extern struct packetdrv drivers[5];	/* 5 packet drivers */
extern int drvnum;		/* 5 drivers are allowed */
extern unsigned int fatal;
extern unsigned int repeater;	/* !=0 for repeating, not bridging */
/*
extern struct pktbuffer recvbuf[PUFFERS];	/* the RxBuffers */
extern int Timer_on;		/* 1=TimerIRQ on */
extern struct ip_adr my_ip;	/* our (bridge) IP address */
extern int do_arp;		/* answer to ARP reply */
extern int tftp_port;
extern unsigned long ping_cnt;	/* amount of ping replys */
extern unsigned long arp_cnt;	/* amount of  ARP replys */
extern unsigned long tftp_cnt;	/* amount of tftp answers */

/* Compute the checksum of an icmp or udp packet,
 * this is inline assembler because it's really too slow otherwise
 */

unsigned int compute_checksum16(char far * pnt, int len)
{
    unsigned int erg, s, o;

    o = FP_OFF(pnt);
    s = FP_SEG(pnt);

    asm {
	push ds
	 mov ax, s		/* set far pointer to data to ds:bx */
	 mov bx, o
	 mov cx, len
	 mov ds, ax
	 mov ax, 0
	 clc
    }
    ch_loop:asm {
	adc ax,[bx]		/* just add words */
	inc bx			/* we're adding words */
	 inc bx			/* has to be word count */
	 dec cx
	 jnz ch_loop
	 adc ax, 0		/* add last carry */
	 not ax			/* invert it */
	 pop ds
	 mov erg, ax		/* return it */
    }
    return erg;
}

/* Some functions necessary to order bytes correctly from host to network
 * order or vice versa
 */
unsigned int ntohs(unsigned int net)
{
    return ((net >> 8) + ((net & 255) << 8));
}

unsigned int htons(unsigned int net)
{
    return ((net >> 8) + ((net & 255) << 8));
}

/* a packet with our ethernet address as destination was received
 * Some tests are made ...
 * 1- is it an ICMP echo request message?
 *    - we send an ICMP echo reply message back
 * 2- is it an udp packet?
 *    - we check for port tftp
 *       * if it's an tftp request we process it
 *          - establish connect
 *          - Set reply port correct: Do not send port unreachable
 *            on our special reply port.
 *       * otherwise ICMP port unreachable
 * 3- is it an tcp packet?
 *    - send a `reset connection' (NOT YET)
 * 4- all other packets:
 *    - ICMP protocol unreachable
 */

int process_pkt(int intno, char far * buffer, unsigned length)
{
    struct ip_packet *pkt = (struct ip_packet *) buffer;

/* First of all we've to process the IP part of the packet
 *
 *    process_ip
 */
    if (pkt->ethtype != htons(0x0800)) {	/* is it an IP-Packet? */
	printf("Notice: Not an IP packet\n");
	return;
    }
    if (compute_checksum16(buffer + 14, 10)) {	/* checksum correct? */
	printf("Error: IP checksum wrong\n");
	return;
    }
    if (((pkt->versionIHL) & 0x0f) != 5) {	/* contains header options? */
	printf("Notice: IP-header contains options, cannot process\n");
	return;
    }
/* Check for protocol, at the moment only icmp, udp, tcp are
 * detected, all others are `protocol unreachable'
 */
    if (pkt->protocol == ICMP) {/* ICMP ? */
	process_icmp(intno, pkt, length);
    } else if (pkt->protocol == UDP) {	/* UDP ?  */
	process_udp(intno, pkt);
    } else {
	printf("Error: Unsupported protocol %d\n", pkt->protocol);

/* We received an unknown protocol type
 * send a `protocol unreachable' message
 */
	do_icmp_proto_unreachable(intno, buffer);
    }
    return;
}

/* send a `protocol unreachable' message via ICMP
 *
 */
void do_icmp_proto_unreachable(int intno, char far * buffer)
{

    struct ip_packet *pkt = (struct ip_packet *) buffer;
    struct icmp *icmp_part = (struct icmp *) pkt->data;

    memmove( (pkt->data) + 8, buffer + 14, 28);
    setup_header(pkt, ICMP, 54);
    icmp_part->type = 3;	/* "Destination unreachable" */
    icmp_part->code = 2;	/* protocol unreachable rfc0792 */
    icmp_part->id = icmp_part->sequence = 0;
    icmp_part->checksum = 0;	/* clear checksum */
    icmp_part->checksum = compute_checksum16(pkt->data, 34 / 2);
    send_pkt(intno, buffer, 68);
    return;
}

/* send a `port unreachable' message via ICMP
 *
 */
void do_icmp_port_unreachable(int intno, char far * buffer)
{

    struct ip_packet *pkt = (struct ip_packet *) buffer;
    struct icmp *icmp_part = (struct icmp *) pkt->data;

    memmove( (pkt->data) + 8, buffer + 14, 28);
    setup_header(pkt, ICMP, 54);
    icmp_part->type = 3;	/* "Destination unreachable" */
    icmp_part->code = 3;	/* port unreachable rfc0792 */
    icmp_part->id = icmp_part->sequence = 0;
    icmp_part->checksum = 0;	/* clear checksum */
    icmp_part->checksum = compute_checksum16(pkt->data, 34 / 2);
    send_pkt(intno, buffer, 68);
    return;
}

/* Setup reply packets
 */
void setup_header(struct ip_packet *pkt, int protocol, int totlen)
{
    memcpy(pkt->destadr, pkt->srcadr, 6);
    memcpy(pkt->srcadr, Ethernetad, 6);
    pkt->ip_dest = pkt->ip_src;
    pkt->ip_src = my_ip;
    pkt->ttl = MAX_TTL;		/* User definable in yabri.h */
    pkt->totlen[0] = totlen >> 8;
    pkt->totlen[1] = totlen & 255;
    pkt->protocol = protocol;
    /* Is calculation necessary? */
    pkt->head_check = 0;	/* clear checksum for calculation */
    pkt->head_check = compute_checksum16((char *) pkt + 14, 10);
}

/* process incoming ICMP packets
 */
void process_icmp(int intno, struct ip_packet *pkt, unsigned length)
{
    struct icmp *echo = (struct icmp *) pkt->data;
    unsigned int len;

    if (echo->type != 8) {	/* ICMP echo request ?
				  if not, discard packet */
	printf("Notice: no ICMP-echo-request\n");
	return;
    }
    len = pkt->totlen[0] * 256 + pkt->totlen[1];
    if (compute_checksum16(pkt->data, (len - 20) / 2)) {
	printf("Error: Echo-checksum wrong\n");
	return;
    }
#ifdef debug_ping
    printf("Echo request received from %s -> sending echo reply\n",
	   sprint_IP(pkt->ip_src));
#endif
    setup_header(pkt, ICMP, len);
    echo->type = 0;
    echo->checksum = 0;		/* clear checksum */
    echo->checksum = compute_checksum16(pkt->data, (len - 20) / 2);
    send_pkt(intno, (char far *) pkt, length);
    ping_cnt++;
}

/* The big tftpd part, tftp is described in rfc0783
 * process the incoming udp packets,
 * - check for port 69 or our `magic port' number
 *    * not detected, send ICMP port unreachable
 * - check for tftp-type
 *    * RREQ (Read requests) establish buffer, start transmission
 *    * WREQ (Write request) access special features ...
 *    *      described further (but always send special error-msg's)
 *    * DATA ACK (Data acknowleged) Send next data packet..
 *    * ERR  Discard silently (must, see rfc0783)
 *    Comment: only one connection is possible ...
 */
void process_udp(int intno, struct ip_packet *pkt)
{
    struct tftp_packet *tftp = (struct tftp_packet *) pkt->data;
    int write_ok=0;
    if (ntohs(tftp->dest_port) == 69) {
	if (ntohs(tftp->opcode) == 1) {
#ifdef DEBUG_TFTP
	    printf("GET file %s ", tftp->u1.filename);
#endif
	    process_rreq(intno, pkt);
	    tftp_cnt++;
	} else if ((ntohs(tftp->opcode) != 5) & \
		   (ntohs(tftp->opcode) != 2)) {
	    send_tftp_error(intno, pkt, 2,"Access denied\n\0");	/* access violation */
	} else if (ntohs(tftp->opcode) == 2) {
	  /* Write Request... */
#ifdef DEBUG_TFTP
	  printf("WRITE file %s ",tftp->u1.filename);
#endif
	  write_ok=0;
	  if (strncmp(tftp->u1.filename,"repeater",8)==0) {
		     repeater=1;
		     write_ok=1;
	  }
	  if (strncmp(tftp->u1.filename,"bridge",6)==0) {
		     repeater=0;
		     write_ok=1;
	  }
	if (write_ok!=0) send_tftp_error(intno, pkt, 0, "Mode set successfull\n\0");
	   else send_tftp_error(intno, pkt, 2, "Access violation\n\0");
	}
	   /*  else dest_port */
    } else if (ntohs(tftp->dest_port) == tftp_port && \
	       conn_established) {
	if (ntohs(tftp->opcode) == 4) {	/* ACK packet */
	    /* packets not in sequence: error: `unknown transfer id' */
	    if (ntohs(tftp->u1.blocknumber) != blocknumber) {
		blocknumber = 1;
		conn_established = 0;	/* reset data transfer */
		send_tftp_error(intno, pkt, 5,"Unexpected ACK packet\n\0");
	    }
	    blocknumber++;
	    if (conn_established != 2) {
		send_tftp_data_block(intno, pkt, output);
	    } else
		conn_established = 0;
	} else if (ntohs(tftp->opcode) != 5) {
	    send_tftp_error(intno, pkt, 2, "Access denied\n\0");	/* access violation */
	}			/* else dest_port */
    } else {			/* wrong port */
	printf("udp: Unsupported port %d\n", ntohs(tftp->dest_port));
	do_icmp_port_unreachable(intno, (char far*) pkt);
    }
}

/* Process tftp RREQ Read request
 * various `files' can be requested
 */
void process_rreq(int intno, struct ip_packet *pkt)
{

    struct tftp_packet *tftp = (struct tftp_packet *) pkt->data;

    output = messagebuffer;     /* Assign output buffer   */
    blocknumber = 1;		/* reset the blocknumber  */
    conn_established = 1;	/* Connection established */
    if (stricmp(tftp->u1.filename, "statistics") == 0) {
	print_stat(output);
	send_tftp_data_block(intno, pkt, output);
    } else if (stricmp(tftp->u1.filename, "info") == 0) {
	print_info(output);
	send_tftp_data_block(intno, pkt,output);
    } else if (stricmp(tftp->u1.filename, "table") == 0) {
	print_table(output);
	send_tftp_data_block(intno, pkt, output);
    } else if (stricmp(tftp->u1.filename, "help") == 0) {
	sprintf(output, "%s\n%s\n%s", yabri, version, help);
	send_tftp_data_block(intno, pkt, output);
    } else {
	send_tftp_error(intno, pkt, 1, "Unknown file, try info table statistics or help\n\0");	/* file not found */
#ifdef DEBUG_TFTP
	printf(" - not found");
#endif
    }
}


void send_tftp_error(int intno, struct ip_packet *pkt, int errnum,\
		     char *errstr)
{
    struct tftp_packet *tftp = (struct tftp_packet *) pkt->data;


    tftp->dest_port = tftp->src_port;
    tftp->src_port = htons(tftp_port);
    tftp->opcode = htons(5);	/* ERROR */
    tftp->u1.errorcode = ntohs(errnum);
    if (errstr != NULL)
      strncpy (tftp->u2.errmsg,errstr,100);
    else
      strcpy(tftp->u2.errmsg,"Unknown error\0");
    setup_header(pkt, UDP, 32+2+strlen(tftp->u2.errmsg));
    tftp->length = htons(12+strlen(tftp->u2.errmsg));
    tftp->checksum = 0;
/* Don't caluculate checksum */
#if 0
    tftp->checksum = compute_checksum16(tftp, 14 / 2);
#endif
    send_pkt(intno, (char far *) pkt, 18 + 20 + 10+strlen(tftp->u2.errmsg));
}

/* Send a `generic' tftp datablock
 * increasing the blocknumber and the ACK packet is handled
 * by process_udp
 */
void send_tftp_data_block(int intno, struct ip_packet *pkt, char *datablock)
{
    unsigned int len;
    struct tftp_packet *tftp = (struct tftp_packet *) pkt->data;
    char *blockpointer;

    tftp->dest_port = tftp->src_port;
    tftp->src_port = htons(tftp_port);
    tftp->opcode = htons(3);	/* DATA */
    tftp->u1.blocknumber = htons(blocknumber);
    if ((blocknumber - 1) * 512 >= strlen(datablock)) {
	printf("send_tftp_data_block: internal error: blocknumber too large");
	send_tftp_error(intno, pkt, 0, "Transfer error");	/* send unknown error */
	return;
    }
    blockpointer = (char *) (datablock + ((blocknumber - 1) * 512));
    len = strlen(blockpointer) + 12;
    if (len > 511 + 12) {
	len = 512 + 12;
    } else if (len & 1) {
	tftp->u2.data[len - 12] = 0;
	len++;
    }
    if (len != 512 + 12)
	conn_established = 2;	/* "FIN", last block send */
    tftp->length = ntohs(len);
    setup_header(pkt, UDP, 20 + len);
    tftp->checksum = 0;
    memcpy(tftp->u2.data,blockpointer,len);
#if 0
    tftp->checksum = compute_checksum16(tftp, len / 2);
#endif
    send_pkt(intno, (char far *) pkt, 12 + 20 + len+2);
}

/* End of file reply.c */

