/* YABRI - Yet Another Bridge for Ethernet */
/* (c) Rainer Bawidamann, Carsten Gross 1995 */

/* yabri.h
   Version 0.1 */

/* History
 *
 * 9.7.95
 *    new struct ip_packet
 * 5.8.95
 *    new struct arp_packet, ip_adr
 * 21.9.95
 *    switch to english
 * 
 * 13.5.96
 *    added user options
 *    MAX_TTL  maximum time_to_live in answers
 *    (heck, sense makes a value of 1 or 2 or crazy: 64)
 *
 */


#define MAX_TTL 64


/* No User definable options below this point ;-) */

#define PUFFERS 20             /* number of packet buffers */
#define TAB_ENTRIES 255        /* size of the lookup table */
#define NAMESIZE 32            /* max size of a driver name */

#define FALSE 0
#define TRUE 1

/* IP protocols */

#define ICMP 1
#define TCP 6
#define UDP 17

/* this struct holds the info about a installed packet driver */
struct packetdrv {
   int intno;		       /* the interrupt vector */
   int oldmode;
   char myeaddr[6];            /* ethernet address of this card */
   int handle;                 /* the handle (not used) */
   int can_as;                 /* driver is able to do as_sendpkt */
   char far name[NAMESIZE];    /* the name of the driver */
};

/* this structure holds an IP address */
struct ip_adr {
   unsigned char adr[4];       /* just 4 bytes */
};

/* this is the format of a ARP packet */
struct arp_packet {
            char destadr[6];   /* destination hardware address */
            char srcadr[6];    /* source hardware address */
   unsigned int ethtype;       /* the type, here 0x0608 */
   unsigned int hw_type;       /* hardware type, only ethernet (1) ok */
   unsigned int prot_type;     /* protocol type, only IP (0800) ok */
   unsigned char hw_len;       /* length of a hardware address (4) */
   unsigned char prot_len;     /* length of a protocol address (6) */
   unsigned int opcode;        /* 1=request, 2=reply */
   unsigned char hw_send[6];   /* hardware address sender */
   struct ip_adr prot_send;    /* protocol address sender */
   unsigned char hw_dest[6];   /* hardware address target */
   struct ip_adr prot_dest;    /* protocol address target */
};

/* this is the format of an IP packet */
struct ip_packet {
            char destadr[6];   /* Okay, okay has nothing to do with ip */
            char srcadr[6];    /* but compatibility issues requires it */
   unsigned int ethtype;
   unsigned char versionIHL;   /* IP version + IHL */
   unsigned char iptype;       /* type of the packet */
   unsigned char totlen[2];    /* length header+data */
                               /* network order!!! */
   unsigned int id;
   unsigned int flag;          /* fragmentation ... */
   unsigned char ttl;          /* Time-To-Live */
   unsigned char protocol;     /* UDP, TCP, ... */
   unsigned int head_check;    /* checksum for the header */
   struct ip_adr ip_src;       /* source protocol address */
   struct ip_adr ip_dest;      /* destination protocol address */
            char data[1480];   /* the restof the packet */
};

/* structure of an ICMP packet (in ip_packet.data) */
struct icmp {
   unsigned char type;         /* ICMP type */
   unsigned char code;         /* 'subtype' */
   unsigned int checksum;      /* checksum of this part */
   unsigned int id;            /* an id for this packet (UDP!) */
   unsigned int sequence;      /* a serial number */
};

/* structure of a TFTP packet (in ip_packet.data) */
struct tftp_packet {
   unsigned int src_port;
   unsigned int dest_port;
   unsigned int length;
   unsigned int checksum;

   unsigned int opcode;
   union {
      char filename[2];
      unsigned int blocknumber;
      unsigned int errorcode;
   } u1;
   union {
      char data[512];
      char errmsg[20];
   } u2;
};

/* structure of a TCP packet (in ip_packet.data) 
 * Note: this is at the moment only necessary for connection reset 
 */
struct tcp_packet {
   unsigned int source;
   unsigned int dest;
   unsigned long seq;
   unsigned long ack;
   unsigned int  res1:4,
                 doff:4,
                 fin:1,
                 syn:1,
                 rst:1,
                 psh:1,
                 ackn:1,
                 urg:1,
                 res2:2;
   unsigned int  window;
   unsigned int  check;
   unsigned int  urg_ptr;
};



/* the buffer to hold a received packet */
struct pktbuffer {
   char buffer[1514];          /* the packet */
   unsigned int TxFlag;        /* what to do with this */
   int drvr;                   /* where the packet came from */
   unsigned int pkt_len;       /* length of the packet */
};

/* an entry in the lookup table */
struct adr_table {
   char addr[6];               /* the ethernet address */
   int drvr;                   /* the line/driver */
   int timeout;                /* unused */
};

/* from yabri.c and reply.c */

extern struct packetdrv drivers[5];       /* the drivers */
extern int drvnum;                        /* number of drivers */
extern unsigned int fatal;
extern unsigned int repeater;             /* !=0 -> repeater mode */
extern struct pktbuffer recvbuf[PUFFERS]; /* the buffers */
extern int Timer_on;                      /* 1=TimerIRQ on */
extern struct ip_adr my_ip;               /* IP address of the bridge */
extern int do_arp;                        /* =1 -> respond to ARP reply */
extern int blocknumber;			  /* actual Blocknumber */
extern int conn_established;		  /* connection established? */
extern void print_drivers(char *);        /* Print driver assignments */
extern void print_info(char *);           /* Print informations       */
extern void print_stat(char *);           /* Print packet driver stat */
extern void print_table(char *);          /* Print address table      */





extern char *output;			  /* Pointer to messagebuf */
extern char messagebuf[2000];		  /* Messagebuffer */
extern char *yabri;			  /* Yabri (C) */
extern char *help;			  /* Helptext  */
extern char *version;			  /* Versionsnummer */



/* from reply.c */

extern void setup_header(struct ip_packet *pkt, int protocol, int totlen);
extern void process_udp(int intno, struct ip_packet *pkt);
extern void process_icmp(int intno, struct ip_packet *pkt, unsigned length);
extern void send_tftp_error(int intno, struct ip_packet *pkt, int errnum, char *errstr);
extern void process_tcp(int intno, struct ip_packet *pkt, unsigned int length);
extern void send_tftp_data_block(int intno, struct ip_packet *pkt, char *datablock);
extern void process_rreq(int intno, struct ip_packet *pkt);
extern void do_icmp_port_unreachable(int intno, char far * buffer);
extern void do_icmp_proto_unreachable(int intno, char far * buffer);
extern int process_pkt(int intno, char far *buffer, unsigned length);
extern unsigned int ntoh(unsigned int);
extern unsigned int hton(unsigned int);



/* from receiver.asm */

extern struct adr_table lookup[TAB_ENTRIES]; /* the lookup table */
/* extern struct adr_table *end_lookup; not used any more */
extern unsigned int adr_cnt;            /* number of addresses in table */
extern char Ethernetad[6];              /* ip address of the bridge */
extern unsigned int screen;             /* address of screen memory */
extern void far receiver1(void);        /* receiver routines for the */
extern void far receiver2(void);        /* drivers */
extern void far receiver3(void);
extern void far receiver4(void);
extern void far receiver5(void);
extern void far flush_table(void);      /* clear lookup table */
extern void far get_TimerIRQ(void);     /* install timer handler */
extern void far release_TimerIRQ(void); /* de-install timer handler */

